body {
    font-family:Arial, sans-serif;
    background-color:White;
    color:Black;
}
hr {
    color:Black;
    background-color:Black;
    height:1px;
    border:0;
}
h1 {
    margin-bottom:0px;
    padding-bottom:0px;
    font-size:1.625em;
}
h2 {
    margin-bottom:0px;
    padding-bottom:0px;
    font-size:1.25em;
    white-space:nowrap;
}
span.section {
    font-weight:bold;
    cursor:pointer;
    color:#003E69;
    white-space:nowrap;
    margin-left:0.3125em;
}
span.new_sign {
    font-weight:bold;
    margin-left:1.65em;
    color:#003E69;
}
span.new_sign_lbl {
    margin-left:3em;
    font-size:1em;
    color:Black;
}
span:hover.section {
    color:#336699;
}
span.sect_aff {
    cursor:pointer;
    padding-left:1.55em;
    font-size:0.875em;
    color:#cc3300;
}
span.sect_info {
    cursor:pointer;
    padding-left:1.55em;
    font-size:0.875em;
    color:Black;
}
span.ext {
    font-weight:normal;
}
span.h_name {
    color:#cc3300;
    font-size:0.875em;
    font-weight:bold;
}
div.h_list, div.lib_list {
    font-size:0.94em;
    padding-left:0.4em;
}
span.ns {
    color:#408080;
    font-size:0.94em;
}
span.lib_name {
    color:Green;
    font-size:0.875em;
    font-weight:bold;
}
span.iname {
    font-weight:bold;
    color:#003E69;
    margin-left:0.3125em;
}
span.iname_b {
    font-weight:bold;
}
span.iname_a {
    color:#333333;
    font-weight:bold;
    font-size:0.94em;
}
span.sym_p {
    font-weight:normal;
    white-space:normal;
}
span.sym_pd {
    white-space:normal;
}
span.sym_p span, span.sym_pd span {
    white-space:nowrap;
}
div.affect {
    padding-left:1em;
    padding-bottom:10px;
    font-size:0.87em;
    font-style:italic;
    line-height:0.9em;
}
div.affected {
    padding-left:1.9em;
    padding-top:10px;
}
table.ptable {
    border-collapse:collapse;
    border:1px outset black;
    margin-left:0.95em;
    margin-top:3px;
    margin-bottom:3px;
    width:56.25em;
}
table.ptable td {
    border:1px solid gray;
    padding:3px;
    font-size:0.875em;
    text-align:left;
    vertical-align:top;
    max-width:28em;
    word-wrap:break-word;
}
table.ptable th.pn {
    width:2%;
}
table.ptable th.chg {
    width:47%;
}
table.vtable {
    border-collapse:collapse;
    border:1px outset black;
    margin-left:1.9em;
    margin-top:0.7em;
}
table.vtable td {
    border:1px solid gray;
    padding:3px;
    font-size:0.875em;
    vertical-align:top;
    max-width:450px;
    word-wrap:break-word;
}
table.ptable th, table.vtable th {
    background-color:#eeeeee;
    font-weight:bold;
    color:#333333;
    font-family:Verdana, Arial;
    font-size:0.875em;
    border:1px solid gray;
    text-align:center;
    vertical-align:top;
    white-space:nowrap;
    padding:3px;
}
table.summary {
    border-collapse:collapse;
    border:1px outset black;
}
table.summary th {
    background-color:#eeeeee;
    font-weight:normal;
    text-align:left;
    font-size:0.94em;
    white-space:nowrap;
    border:1px inset gray;
    padding:3px;
}
table.summary td {
    text-align:right;
    white-space:nowrap;
    border:1px inset gray;
    padding:3px 5px 3px 10px;
}
span.mngl {
    padding-left:1em;
    font-size:0.875em;
    cursor:text;
    color:#444444;
    font-weight:bold;
}
span.pleft {
    padding-left:2.5em;
}
span.sym_ver {
    color:#333333;
    white-space:nowrap;
    font-family:"DejaVu Sans Mono", Monospace;
}
span.attr {
    color:#333333;
    font-weight:normal;
}
span.color_p {
    font-style:italic;
    color:Brown;
}
span.p {
    font-style:italic;
}
span.fp {
    font-style:italic;
    background-color:#DCDCDC;
}
span.ttype {
    font-weight:normal;
}
span.nowrap {
    white-space:nowrap;
}
span.value {
    font-weight:bold;
}
.passed {
    background-color:#CCFFCC;
    font-weight:normal;
}
.warning {
    background-color:#F4F4AF;
    font-weight:normal;
}
.failed {
    background-color:#FFCCCC;
    font-weight:normal;
}
.new {
    background-color:#C6DEFF;
    font-weight:normal;
}
.compatible {
    background-color:#CCFFCC;
    font-weight:normal;
}
.almost_compatible {
    background-color:#FFDAA3;
    font-weight:normal;
}
.incompatible {
    background-color:#FFCCCC;
    font-weight:normal;
}
.gray {
    background-color:#DCDCDC;
    font-weight:normal;
}
.top_ref {
    font-size:0.69em;
}
.footer {
    font-size:0.75em;
}
